(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: check.mli,v 1.35 2008/02/11 23:10:05 weis Exp $ *)

(** {3 Builds various hash tables for generators and types.} *)

open Parsetree
;;

(** Possible errors. *)

type error =
   | Unknown_generator of Longident.t
   | No_neutral_element of Longident.t
   | Public_type_with_relations of string
   | Cannot_find_module of Longident.t
   | Not_supported of string
   | Relation_already_declared of Location.t
   | Incompatible_types of Longident.t * Longident.t
   | Invalid_pattern
   | Pattern_not_headed_by_correct_generator of Longident.t
   | Illegal_relation of Longident.t
;;

exception Error of Location.t * error
;;

val error : Location.t -> error -> 'a
;;

val report_error : Format.formatter -> error -> unit
;;

(** The locality flag for a generator: is it defined in the local module or
    globally ? *)
type locality =
   | Local
   | Global of string
;;

(** Information for a generator. *)
type generator_info = {
  gi_type_name : string;
  gi_locality : locality;
  gi_arg_types : core_type list;
  gi_relations : relations;
  gi_location : Location.t;
  gi_priv_flag : Asttypes.private_flag;
}
;;

(** Various access/test functions. *)

val is_local_info : generator_info -> bool
;;
val is_private_info : generator_info -> bool
;;
val is_varyadic_info : generator_info -> bool
;;
val arity_info : generator_info -> int
;;

(** Enters a list of type definitions in the table. *)
val add_type_decls : locality -> (string * type_declaration) list -> unit
;;

(** Various access functions that may raise Not_found. *)

val find_gident : Longident.t -> generator_info
;;
val find_gname : string -> generator_info
;;
val find_tident : Longident.t -> type_declaration
;;
val find_tname : string -> type_declaration
;;

(** Various access/test functions that may raise Unknown_generator. *)

val is_varyadic_gen : Parsetree.generator -> bool
;;
val is_local_gen : Parsetree.generator -> bool
;;
val is_private_gen : Parsetree.generator -> bool
;;
val arity_gen : Parsetree.generator -> int
;;
val is_truly_unary_gen : Parsetree.generator -> bool
;;

val neutral_element_gen : Parsetree.generator -> Parsetree.generator
(** Gives the neutral element of [g] or else raise No_neutral_element *)
;;
