(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: comp.ml,v 1.14 2008/02/12 12:08:44 weis Exp $ *)

open Cp;;
open Equation;;
open Rule;;
open Norm;;
open Rename;;
open Subs;;
open Match;;

type ers = EqnSet.t * RulSet.t;;

(* combinators for applying completion rules *)

let bprint b (es,rs) =
  if !Debug.print_rules then
    Printf.bprintf b "equations: %a;\nrules: %a\n"
      EqnSet.bprint es RulSet.bprint rs;
  if !Debug.print_stats then
    Printf.bprintf b "equations: %d; rules: %d\n"
      (EqnSet.cardinal es) (RulSet.cardinal rs);;

let apply s rule x =
  Debug.output_string "rule" (s ^ " begin");
  let y = rule x in
    Debug.output "rule" (s ^ " end") bprint y;
    y;;

let apply_opt_once s rule x =
  Debug.output_string "rule" (s ^ " begin");
  match rule x with
    | Some y -> Debug.output "rule" (s ^ " end") bprint y; y
    | None -> Debug.output_string "rule" (s ^ " end"); x;;

let repeat rule =
  let rec aux ers =
    match rule ers with
      | Some ers -> aux ers
      | _ -> ers
  in aux;;

let repeat s rule ers =
  Debug.output_string "rule" (s ^ " begin");
  let ers = repeat rule ers in Debug.output "rule" (s ^ " end") bprint ers; ers;;

(* rule delete. delete all trivial equations l=l at once *)

(*let delete =
  let f (l,r) = l <> r in
    fun (es,rs) -> EqnSet.filter f es, rs;;

let delete = apply "delete" delete;;*)

(* rule simplifity. simplify all equations at once *)

(*let simplify (es,rs) =
  let f (l,r) = (norm rs l, norm rs r) in
    EqnSet.map f es, rs;;

let simplify = apply "simplify" simplify;;*)

(* combination of simplify and delete *)

(*let del_simp ers = delete (simplify ers);;*)

let del_simp (es,rs) = remove_trivial (norm rs) es, rs;;

let del_simp = apply "del_simp" del_simp;;

(* rule orient. orient all possible equations at once.
raise Not_found if there is no orientable equation *)

type 'a gt = 'a -> 'a -> bool;;

exception Unorientable of EqnSet.t;;

let add_rev (l,r) = RulSet.add (r,l);;

let orient gt (es,rs) =
  let esgt, es2 = EqnSet.partition (fun (l,r) -> gt l r) es in
  let eslt, es2 = EqnSet.partition (fun (l,r) -> gt r l) es2 in
    if EqnSet.is_empty esgt && EqnSet.is_empty eslt
    then raise (Unorientable es)
    else (es2, EqnSet.fold RulSet.add esgt (EqnSet.fold add_rev eslt rs));;

let orient gt = apply "orient" (orient gt);;

(* rule compose: try to reduce a rhs *)

let compose (es,rs) =
  let f (l,r) = (l, norm rs r) in
    es, RulSet.map f rs;;

let compose = apply "compose" compose;;

(* test whether a rule is strictly greater than another rule *)
 
let rule_gt gt (l1,r1) (l2,r2) =
  match Rename.eq_opt l1 l2 with
    | Some s -> gt (Subs.apply s r1) r2
    | _ -> insert_gt l1 l2;;

(* rule collapse: try to reduce a lhs *)

let collapse_one gt (es,rs) =
  let p ((l,r) as lr) =
    let rs2 = RulSet.filter (rule_gt gt lr) rs in
      match red_opt rs2 l with
	| Some l2 -> Some (EqnSet.add (l2,r) es, RulSet.remove lr rs)
	| _ -> None
  in RulSet.find_opt p rs;;

let collapse gt = repeat "collapse" (collapse_one gt);;

(*let collapse_one gt = apply_opt_once "collapse" (collapse_one gt);;*)

(* rule deduce *)

let deduce acs (es,rs) = add_cps (norm rs) es (RulSet.union acs rs), rs;;

let deduce acs = apply "deduce" (deduce acs);;

(* completion procedure *)

let step acs gt ers =
  deduce acs (del_simp (collapse gt (compose (orient gt ers))));;

let complete_n acs gt n =
  let rec aux i ((es,_) as ers) =
    if EqnSet.is_empty es || i=n then ers,i
    else (
      Debug.output_string_sep "step" (string_of_int (i+1) ^ " begin");
      let ers = step acs gt ers in
	Debug.output "step" (string_of_int (i+1) ^ " end") bprint ers;
	aux (i+1) ers)
  in aux 0;;

let complete acs gt = complete_n acs gt 0;;
