(* $Id: listutils.ml,v 1.4 2008/01/09 14:02:18 lowenthl Exp $ *)

(** {Miscelaneous list functions} *)

open List
open Funutils
;;

(** Returns true iff list is empty *)
let null l = match l with
  [] -> true
| _ -> false
;;

(** Adds an element to the beggining of a list *)
let cons x xs = x :: xs
;;

(** Takes first n elements of a list, or the whole list if it has less than n *)
 (** elements *)
let rec take n l = match n, l with
  0, _  -> []
| _, [] -> []
| n, x::xs -> x :: take (n-1) xs
;;

(** Drops first n elements of a list, or the whole list if it has less than n *)
 (** elements *)
let rec drop n l = match n, l with
  0, l  -> l
| _, [] -> []
| n, _::xs -> drop (n-1) xs
;;

(** Takes the sublist starting from position n and ending in position m   (numbering from 0). Cuts the sublist accordingly if there are not enough elements. *)
let sublist n m = take (m - n + 1) <.> drop n
;;

(** Separates the list in consecutive groups of the specified size, then returns the i_th of such groups starting from 0. If there are not enough elements, cuts the group accordingly. *)
let ith_group i group_size =
  let init, last = group_size * i, group_size * (i + 1) - 1 in
  sublist init last

(** map_filter f p [a1; ...; an] applies function f only to those a1, ..., an *)
 (** for which predicate p holds, and builds the list [f ai1; ...; f aik]  with the results returned by f.*)
let rec map_filter f p l = match l with
  [] -> []
| x::xs -> if p x then f x :: map_filter f p xs else map_filter f p xs

let flat_map_end f l accu =
  List.rev (
   List.rev_append accu
     (List.fold_left
       (fun ll l -> List.rev_append (f l) ll)
       []
       l)
  )
;;

(** flat_map f [a1 ... an] applies function f, which returns a list, to *)
 (** a1,...,an. Then flattens all the lists obtained this way and returns *)
  (** the resulting list. *)
let flat_map f l = flat_map_end f l [];;


(** Maps a function to a list of elements, provided the start, the end and an *)
 (** incrementing function *)
let rec map_from_to f init last inc = 
  if init > last
  then []
  else f init :: map_from_to f (inc init) last inc
;;

(** Creates a list from a starting point to and end, given an incrementing *)
 (** function *)
let from_to init last inc = map_from_to (fun x -> x) init last inc
;;

(** Creates a list containing n times the given element **)
let rec repeat n elem = if n = 0 then [] else elem :: repeat (pred n) elem
;;

(** Transposes a list of lists. *)
 (** For example, transpose [[1,2],[3,4,5],[]] = [[1,3],[2,4],[5]] *)
let rec transpose = function
  | [] ->  []
  | [] :: xss -> transpose xss
  | (x::xs) :: xss -> (x :: map_filter hd (not <.> null) xss) :: 
                      transpose (xs :: map_filter tl (not <.> null) xss)
;;

(** Takes and element of each list to build a new one. Returns all the lists *)
 (** built this way. *)
  (** For example, combine_all [[1,2],[3,4,5],[6]] = [[1,3,6],[1,4,6],[1,5,6],[2,3,6],[2,4,6],[2,5,6]]*)
let rec combine_all lists = match lists with
  [] -> []
| [xs] -> map (fun x -> [x]) xs
| xs :: xss ->
    let xss_combination = combine_all xss in
    flat_map (fun x -> map (cons x) xss_combination) xs


(** Removes an element from a list, provided an equality function  *)
let rec remove_eq eq x = function
  | [] -> []
  | y :: ys when eq x y -> remove_eq eq x ys
  | y :: ys -> y :: remove_eq eq x ys
;;

(** Removes an element from a list, comparing by structural equality *)
let remove x l = remove_eq (fun x y -> x = y) x l
;;

(** Removes duplicates from a list, provided an equality function. Preserves the order of the first appearences. *)
let rec nub_eq eq = function
  | [] -> []
  | x :: xs -> x :: nub_eq eq (remove_eq eq x xs)
;;

(** Removes duplicates from a list, comparing by structural equality. Preserves the order of the first appearences *)
let nub l = nub_eq (fun x y -> x = y) l
;;
